import difflib
import json
from dataclasses import dataclass, field
from enum import Enum

ResultKind = Enum('ResultKind', 'Pass Fail AskServer')
Transition = Enum('Transition', 'Pass Next Fail')
AttemptState = Enum('AttemptState', 'Start Build Fix Change Done')


class ValidationRule:
    def to_data(self):
        raise NotImplementedError()

    @staticmethod
    def from_data(data):
        match data['kind']:
            case 'lines': return LineDiffRule(data['data'])
            case _: raise ValueError(f"unknown rule kind: {data['kind']}")

    @staticmethod
    def from_json(jsn: str) -> 'ValidationRule':
        return ValidationRule.from_data(json.loads(jsn))


class LineDiffRule(ValidationRule):

    def __init__(self, expected: [str]):
        self.expected = expected

    def to_data(self):
        return {'kind': 'lines', 'data': self.expected}


class TestFailure(AssertionError):
    def error_lines(self):
        raise NotImplementedError()

    def print_error(self):
        for line in self.error_lines():
            print(line)


class LineDiffFailure(TestFailure):

    def __init__(self, actual: [str], diff: [str]):
        self.actual = actual
        self.diff = diff

    @staticmethod
    def from_lines(actual: [str], expected: [str]):
        return LineDiffFailure(actual, list(difflib.Differ().compare(actual, expected)))

    @staticmethod
    def from_data(data: dict) -> 'LineDiffFailure':
        return LineDiffFailure(data['actual'], data['diff'])

    def to_data(self):
        return {
            'kind': 'diff',
            'data': {'actual': self.actual, 'diff': self.diff}}

    def error_lines(self):
        return (["---- how to patch your output to pass the test ----"]
                + self.diff)


@dataclass
class TestResult:
    kind: ResultKind
    error: TestFailure = None
    rule: ValidationRule = None
    actual: [str] = field(default_factory=list)

    @staticmethod
    def from_data(data: dict) -> 'TestResult':
        res = TestResult(kind=ResultKind[data['kind']])
        match res.kind:
            case ResultKind.AskServer: raise RecursionError()
            case ResultKind.Fail:
                err = data['error']
                match err['kind']:
                    case 'diff': res.error = LineDiffFailure.from_data(err['data'])
                    case _: raise ValueError(f"unknown error kind: {err['kind']}")
            case ResultKind.Pass:
                rule = data['rule']
                match rule['kind']:
                    case 'lines': res.rule = LineDiffRule(rule['data'])
                    case _: raise ValueError(f"unknown rule kind: {rule['kind']}")
        return res

    def to_data(self):
        """return raw data for json serialization"""
        kind = self.kind.name
        match kind:
            case 'Fail': return {'kind': kind, 'actual': self.actual,
                                 'error': self.error.to_data()}
            case 'Pass': return {'kind': kind, 'rule': self.rule.to_data()}
            case 'AskServer': raise NotImplementedError()

    def is_pass(self):
        return self.kind == ResultKind.Pass


@dataclass
class TestDescription:
    id: int = 0
    chid: int = 0
    grp: int = 0
    ord: int = 0
    name: str = ''
    head: str = ''
    body: str = ''
    ilines: [str] = field(default_factory=list)
    olines: [str] = field(default_factory=list)

    @property
    def rule(self) -> ValidationRule | None:
        if self.olines is None:
            return None
        else:
            return LineDiffRule(self.olines)

    def check_output(self, actual: [str]) -> TestResult:
        if self.olines is None:
            return TestResult(ResultKind.AskServer)
        elif self.olines == actual:
            rule = LineDiffRule(self.olines)
            return TestResult(ResultKind.Pass, rule=rule)
        else:
            e = LineDiffFailure.from_lines(actual=actual, expected=self.olines)
            return TestResult(ResultKind.Fail, error=e, actual=actual)


@dataclass
class Challenge:
    id: int = 0
    sid: int = 0
    name: str = ''
    server: str = ''
    title: str = ''
    tests: [TestDescription] = field(default_factory=list)


DEFAULT_TARGET = './my-program'


@dataclass
class Config:
    uid: int = 0
    program_args: [str] = field(default_factory=lambda: [DEFAULT_TARGET])
    use_shell: bool = False
    # where to write the input to the child program
    # (for cases where stdin is not available)
    input_path: str = ''
    # path to test plan (if running directly from org file)
    test_plan: str = ''
    # skip this many lines of input before sending
    # for cases where the language prints a header
    # that can't be suppressed. (e.g. Godot4 - you can turn
    # off the header, but doing so turns off all prints!)
    skip_lines: int = 0
    attempt: str = ''

    @staticmethod
    def default_target():
        return DEFAULT_TARGET

    def to_json(self):
        data = {
            'attempt': self.attempt,
            'targets': {
                'main': {
                    'args': self.program_args,
                    'shell': self.use_shell}}}
        return json.dumps(data, indent=2)



import asyncio
import inspect
import json
import string
import random

import quart
import jwt as jwtlib

from . import database as db
from . import model as m

app = quart.Quart(__name__)
# TODO: make this dev-server only
app.config['TEMPLATES_AUTO_RELOAD'] = True

THIS_SID = 1  # TODO: validate server id

ok = None

# this maps pre-tokens to async queues that
# will eventually yield the jwt token.
queues: {'pretoken': [asyncio.Queue]} = {}

observers: {'attempt': [asyncio.Queue]} = {}


# == sessions =================================================


def random_string(length=32):
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))


def get_session(skey: str) -> dict | None:
    # TODO: update the timestamp in the 'seen' column
    rows = db.query("select * from sessions where skey=?", [skey])
    return json.loads(rows[0]['data']) if rows else None


def new_session(sid: int, uid: int) -> str:
    skey = random_string()
    db.commit("""
        insert into sessions (skey, sid, uid, data) values (?, ?, ?, ?)
        """, [skey, sid, uid, '{"uid": %i}' % uid])
    return skey


class PleaseLogin(Exception):
    """raised when a request requires a user to be logged in"""
    pass


def require_uid(f0):
    """supplies the uid from the request, or raises PleaseLogin"""
    async def f(*a, **kw):
        # uid can come from cookie or jwt
        skey = quart.request.cookies.get('sess', '')
        if skey:
            uid = (get_session(skey) or {}).get('uid')
        else:
            jsn = await quart.request.json
            if not jsn:
                raise PleaseLogin()
            if not (jwt := jsn.get('jwt')):
                raise LookupError('no jwt given')
            r = db.query('select uid from tokens where jwt=?', [jwt])
            if not r: raise LookupError('unrecognized jwt')
            uid = r[0]['uid']
        if not uid:
            raise PleaseLogin()
        return await f0(uid=uid, *a, **kw)
    f.__name__ = f0.__name__
    return f


@app.errorhandler(PleaseLogin)
async def handle_please_login(_e):
    html = await quart.render_template('please_login.html')
    return html   # htmx didn't show content when status 401


# == websocket notifications ==================================

async def notify(code, data, wrap=True):
    data = f'<span id="test-detail">{data}</span>' if wrap else data
    for o in observers.get(code, []):
        await o.put(data)


async def notify_state(code, state, focus):
    data = dict(state=state, focus=focus, code=code)
    html = await quart.render_template('state.html', data=data)
    await notify(code, html, wrap=False)


JWT_OBJ = jwtlib.JWT()
JWT_KEY = jwtlib.jwk_from_pem(open('tanco_auth_key.pem', 'rb').read())

# == platonic apps ============================================
# mini web framework to wrap htmx fragments in a layout
# and also allow fetching raw json with .json suffix


def htmx_fragment(f0):
    async def f(*a, **kw):
        body = (await f0(*a, **kw)) if inspect.iscoroutinefunction(f0) else f0(*a, **kw)
        qr = quart.request
        if qr.headers.get('HX-Request'):
            return body
        else:
            return await quart.render_template('index.html', body=body)
    f.__name__ = f0.__name__
    return f


def platonic(route, template, hx=True):
    """allows an endpoint that returns data to serve html or json,
    depending on the presence of the string '.json' in the url"""
    def decorator(f0):
        async def fp(*a, **kw):
            data = (await f0(*a, **kw)) if inspect.iscoroutinefunction(f0) else f0(*a, **kw)
            if quart.request.path.endswith('.json'):
                return data
            else:
                return await quart.render_template(template, data=data, url=quart.request.path)

        async def fj(*a, **kw):
            return await fp(*a, **kw)

        @htmx_fragment
        async def fhx(*a, **kw):
            return await fp(*a, **kw)

        fp.__name__ = f0.__name__
        fhx.__name__ = f0.__name__
        fj.__name__ = f0.__name__ + '_json'
        app.route(route)(fhx if hx else fp)
        app.route(route + '.json')(fj)
        return fp
    return decorator

# -------------------------------------------------------------


@app.route('/')
async def index():
    return await quart.render_template('index.html')


@app.route('/about')
@htmx_fragment
async def about():
    return await quart.render_template('about.html')


@platonic('/me', 'me.html')
@require_uid
async def me(uid):
    data = db.query("""
        select u.username
        from users u
        where u.id=?
        """, [uid])[0]
    data['attempts'] = db.query("""
        select a.ts, c.name as c_name, c.title, a.code
        from attempts a, challenges c, users u
        where a.chid=c.id and u.id=? and a.uid=u.id
        """, [uid])
    return data


@platonic('/c', 'challenges.html')
def list_challenges():
    return db.query("""
        select c.id, c.name, c.title
        from challenges c""")


@platonic('/c/<name>', 'challenge.html')
async def show_challenge(name):
    data = db.query("""
        select c.id, c.name, c.title,
          (select count(*) from tests t
           where t.chid = c.rowid) as num_tests
        from challenges c
        where name=?
        """, [name])
    quart.abort(404) if not data else ok
    return data[0]


@app.route('/c/<name>/attempt', methods=['POST'])
@require_uid
async def attempt_challenge(name, uid):
    try:
        [row] = db.query('select id from challenges where name=?', [name])
        chid = row['id']
    except ValueError:
        raise LookupError(f'invalid challenge: {name!r}')
    code = random_string()
    db.commit("""
        insert into attempts (uid, chid, code) values (?, ?, ?)
        """, [uid, chid, code])
    return {'aid': code}


@platonic('/a/<code>', 'attempt.html')
@require_uid
async def show_attempt(code, uid):
    # TODO: trap IndexError if no attempt found
    data = db.query("""
        select a.code, a.state, t.name as focus, c.name as c_name, u.username as u_name
        from challenges c, users u, attempts a left join tests t on a.focus = t.id
        where a.code = (:code) and u.id = (:uid)
        """, {'code': code, 'uid': uid})[0]
    data['state'] = m.AttemptState[data['state'].capitalize()]
    data['progress'] = db.query("""
        select t.name as t_name, p.ts from attempts a, tests t, progress p
        where a.code = (:code) and a.id = p.aid and p.tid = t.id
        """, [code])
    return data


@app.websocket('/a/<code>/live')
# TODO: @require_uid (raises RuntimeError: Not within a request context)
async def attempt_live(code):
    ws = quart.websocket
    q = asyncio.Queue()
    global observers
    observers.setdefault(code, []).append(q)
    try:
        await ws.send('')
        while True:
            html = await q.get()
            await ws.send(html)
    except asyncio.CancelledError:
        observers[code].remove(q)


@platonic('/a/<code>/t/<name>', 'test.html')
@require_uid
async def show_test(**kw):
    # TODO: make sure the user has either passed the test or it is their next test
    data = db.query("""
        select t.name, t.head, t.body, t.ilines,
           t.olines
        from attempts a, tests t        
        where a.chid = t.chid and a.uid = (:uid)
          and a.code = (:code) and t.name=(:name)
        """, kw)[0]
    return data


@app.route('/a/<code>/next', methods=['POST'])
@require_uid
async def next_tests_for_attempt(code, uid):
    db.set_attempt_state(uid, code, m.Transition.Next)
    rows = db.get_next_tests(code, uid)
    # hide the answers for now:
    for row in rows:
        row['olines'] = None
    return rows


@app.route('/a/<code>/pass', methods=['POST'])
@require_uid
async def send_attempt_pass(code, uid):
    state, focus = db.set_attempt_state(uid, code, m.Transition.Pass)
    assert not focus, "all tests passed so focus should be empty"
    await notify_state(code, state, focus='')
    await notify(code, await quart.render_template('pass.html'))
    return ['ok']


@app.route('/a/<code>/fail', methods=['POST'])
@require_uid
async def send_attempt_fail(code, uid):
    # TODO: validate the jwt
    jsn = await quart.request.json
    tr_data = jsn.get('result')
    tr = m.TestResult.from_data(tr_data)
    try:
        tn = jsn['test_name']
        t = db.get_attempt_test(uid, code, tn)
    except KeyError:
        return f'unknown test', 400
    except LookupError:
        return "unknown test or attempt", 400
    state, focus = db.set_attempt_state(uid, code, m.Transition.Fail, failing_test=tn)
    await notify_state(code, state, focus)
    html = await quart.render_template('result.html', test=t, result=tr)
    await notify(code, html)
    return ['ok']


@app.route('/a/<code>/check/<test_name>', methods=['POST'])
@require_uid
async def check_test_for_attempt(code, test_name, uid):
    actual = (await quart.request.json).get('actual')
    if actual is None:
        return "bad request: no 'actual' field in post", 400

    # fetch the expected output
    # TODO: update to allow arbitrary validation rules
    try:
        t = db.get_attempt_test(uid, code, test_name)
    except LookupError:
        return "unknown test or attempt", 404

    r = t.check_output(actual)
    print('test result:', r.to_data())

    if r.is_pass():
        db.save_progress(code, test_name, True)

    if obs := observers.get(code, []):
        html = await quart.render_template('result.html', test=t, result=r, actual=actual)
        for o in obs:
            await o.put(html)
    return r.to_data()


# == Website Authentication ===================================

@app.route('/whoami', methods=['GET'])
async def get_whoami():
    skey = quart.request.cookies.get('sess', '')
    uid = (get_session(skey) or {}).get('uid') if skey else None
    data = db.query('select username from users where id=?', [uid])[0] if uid else {}
    return await quart.render_template('whoami.html', uid=uid, data=data)


@app.route('/login', methods=['GET'])
async def get_login():
    return await quart.render_template('login.html')


@app.route('/login/success', methods=['POST'])
async def post_login_success():
    frm = await quart.request.form
    uid, _ = decode_access_token(frm.get('accessToken'))
    key = new_session(THIS_SID, uid)
    whence = frm.get('whence') or '/'  # could be there but blank
    res = quart.redirect(whence)
    res.set_cookie('sess', key)
    return res


# == Authentication for Command Line Client ===================

@app.route('/auth/login', methods=['GET'])
async def get_auth_login():
    data = {'pre': quart.request.args.get('pre', '??')}
    return await quart.render_template('login.html', **data)


@app.route('/auth/pre', methods=['GET'])
def get_auth_pre():
    return """
    <form method="POST" action="/auth/pre">
        <input type="submit" value="get pre-token">
    </form>
    """


@app.route('/auth/pre', methods=['POST'])
def post_auth_pre():
    pre = random_string()
    queues[pre] = asyncio.Queue()
    return {'token': pre}


@app.route('/auth/jwt', methods=['GET'])
def get_auth_jwt():
    return """
    <form method="POST" action="/auth/jwt">
        <label for="pre">pre-token:</label>
        <input type="text" name="pre">
        <input type="submit" value="get jwt-token">
    </form>"""


@app.route('/auth/jwt', methods=['POST'])
async def post_auth_jwt():
    req = quart.request
    pre = (await req.json).get('pre') if req.is_json else (await req.form).get('pre')
    assert pre in queues, f"pre-token not found: {pre}"
    print(f'awaiting jwt for pre[{pre}]:')
    jwt = await queues[pre].get()
    del queues[pre]
    print(f'jwt for pre[{pre}]:', jwt)
    return {'token': jwt}


def decode_access_token(acc0):
    """returns uid, token_data"""
    # TODO: validate acc against auth provider (firebase)
    # (otherwise attacker could just send any token)
    acc = JWT_OBJ.decode(acc0, do_verify=False)
    # TODO: move this to a real signup process
    # TODO: use real usernames
    authid = acc['sub']
    username = acc['email']
    uid = db.uid_from_tokendata(THIS_SID, authid, username)
    token_data = {'authid': authid, 'username': username}
    return uid, token_data


@app.route('/auth/success', methods=['POST'])
async def post_auth_success():
    frm = await quart.request.form
    uid, data = decode_access_token(frm.get('accessToken'))
    # TODO: validate the jwt
    jwt = JWT_OBJ.encode(data, JWT_KEY, alg='RS256')

    db.commit("insert into tokens (uid, jwt) values (?, ?)",
              [uid, jwt])

    # now tell jwt to the listening command line client
    pre = frm.get('preToken')
    try:
        q = queues[pre]
        print("queues[pre]:", q)
        print("jwt:", jwt)
        q.put_nowait(jwt)
    except KeyError:
        return f"pre-token not found: {pre}", 500
    except asyncio.QueueFull:
        return "pre-token already used", 500
    return """
    <h1>Success!</h1>
    <p>You have successfully logged in.</p>
    <p>You can close this browser tab.</p>
    """


if __name__ == '__main__':
    app.run(host='localhost', port=5000)



#!/bin/env python
"""
command-line driver for tanco client.
"""
import os, sys, cmd as cmdlib, jwt as jwtlib
import sqlite3
import webbrowser
import subprocess

from . import runner, orgtest, database as db, model as m
from .client import TancoClient
from .model import Config, TestDescription


class TancoDriver(cmdlib.Cmd):
    prompt = "tanco> "
    completekey = ''
    cmdqueue = ''

    def __init__(self):
        super().__init__()
        self.result = None    # for passing data between commands
        self.client = TancoClient()

    # -- global database state --------------------------------

    def do_login(self, _arg):
        """Login to the server"""
        if who := self.client.whoami():
            print(f"Already logged in to {self.client.url} as {who['username']}.")
            return
        sid = db.get_server_id(self.client.url)
        pre = self.client.get_pre_token()
        webbrowser.open(self.client.url + '/auth/login?pre=' + pre)
        jwt = self.client.get_jwt(pre=pre)
        data = jwtlib.JWT().decode(jwt, do_verify=False)  # TODO: verify
        uid = db.uid_from_tokendata(sid, data['authid'], data['username'])
        db.commit('insert into tokens (uid, jwt) values (?, ?)', [uid, jwt])

    def do_whoami(self, _arg):
        """Show the current user"""
        try:
            who = self.client.whoami()
        except LookupError as e:
            print(e)
            return
        print(f"logged in as {who['username']}" if who else "not logged in.")

    @staticmethod
    def do_delete(arg):
        """Delete a challenge"""
        if not arg:
            print("Usage: `delete <challenge name>`")
            return
        old = db.query('select id from challenges where name=?', [arg])
        if not old:
            print(f'Sorry. Challenge "{arg}" does not exist in the database.')
            return
        old = old[0]['id']
        tx = db.begin()
        tx.execute('delete from tests where chid=?', [old])
        # TODO: tx.execute('delete from progress where chid=?', [old])
        tx.execute('delete from challenges where id=?', [old])
        tx.commit()
        print(f'Challenge "{arg}" deleted.')

    @staticmethod
    def do_import(arg):
        """Import a challenge"""
        if not arg:
            print("usage: import <challenge.org>")
            return
        if os.path.exists(arg):
            c = orgtest.read_challenge(arg)
            if not (sids := db.query('select id from servers where url=?', [c.server])):
                print(f'Sorry, server "{c.server}" is not in the database.')
                return
            sid = sids[0]['id']
            if db.query('select * from challenges where sid=? and name=?', [sid, c.name]):
                print(f'Sorry, challenge "{c.name}" already exists in the database.')
                print(f'Use `tanco delete {c.name}` if you want to replace it.')
                return
            tx = db.begin()
            cur = tx.execute('insert into challenges (sid, name, title) values (?, ?, ?)',
                             [sid, c.name, c.title])
            chid = cur.lastrowid
            for (i, t) in enumerate(c.tests):
                tx.execute("""
                    insert into tests (chid, name, head, body, grp, ilines, olines)
                    values (?, ?, ?, ?, ?, ?, ?)
                    """, [chid, t.name, t.head, t.body, i,
                          '\n'.join(t.ilines), '\n'.join(t.olines)])
            tx.commit()
            print(f'Challenge "{c.name}" imported with {len(c.tests)} tests.')

    def do_challenges(self, _arg):
        """List challenges"""
        print(f'Listing challenges from {self.client.url}')
        print()
        self.result = self.client.list_challenges()
        for c in self.result:
            print(f' {c["name"]:16} : {c["title"]}')
        print()

    # -- local project config ---------------------------------

    def do_init(self, arg):
        """Create .tanco file in current directory"""
        if not (who := self.client.whoami()):
            print("Please login first.")
            return

        if os.path.exists('.tanco'):
            cfg = runner.load_config()
            if cfg.attempt:
                print("Already initialized.")
                print("Remove the 'attempt' field from .tanco if you")
                print("really want to run `tanco init` again.")
                return
        else:
            cfg = Config()

        if arg:
            self.result = self.client.list_challenges()
        else:
            self.do_challenges('')
        while not arg:
            print("Enter the name of the challenge you want to work on.")
            arg = input("> ")
        match = [c for c in self.result if c['name'] == arg]
        if not match:
            print(f"Sorry, challenge '{arg}' not found.")
            return
        c = match[0]
        sid = db.get_server_id(self.client.url)
        db.commit("""insert or ignore into challenges
            (sid, name, title) values (?, ?, ?)
            """, [sid, c['name'], c['title']])
        chid = db.query('select id from challenges where name=?', [c['name']])[0]['id']
        # now we have arg = a valid challenge name on the server,
        # so we have to initialize the attempt on both the remote
        # and local databases.
        code = self.client.attempt(arg)
        uid = who['id']
        db.commit("""
            insert into attempts (uid, chid, code) values (?, ?, ?)
            """, [uid, chid, code])
        cfg.attempt = code
        with open('.tanco', 'w') as f:
            f.write(cfg.to_json())
        print('Project initialized.')
        print('Edit .tanco to configure how to run your project.')
        print('Then run `tanco check` to make sure tanco can run your program.')

    @staticmethod
    def do_check(arg):
        """Check that the target program runs."""
        runner.check(['tanco']+[x for x in arg.split(' ') if x != ''])

    def do_show(self, arg=None):
        """Show a description of the current test."""
        cfg = runner.load_config()
        state = db.current_state(cfg.attempt)
        if state == 'start':
            print("You have not started the challenge yet.")
            print("Use `tanco check` to make sure your program runs.")
            print("Use `tanco next` to fetch the first test.")
            return
        tests = db.get_next_tests(cfg.attempt, cfg.uid)
        self.result = (cfg.attempt, tests)
        if tests:
            if arg == '-n':
                print('You already have the next test. Calling `tanco show`:')
                print()
            t = TestDescription(**tests[0])
            print(f'#[{t.name}]: {t.head}')
            print()
            print(t.body)
            print()
            print("Use 'tanco test' to run the tests.")
        elif arg == '-n':
            pass
        else:
            print("All known tests have passed.")
            print("Use `tanco test` to check that they still pass.")
            print("Use `tanco next` to fetch the next test.")

    @staticmethod
    def do_version(_arg):
        """Print tanco version (by calling `pip show tanco`) """
        subprocess.run(["pip", "show", "tanco"])

    @staticmethod
    def do_status(_arg):
        """print information about the current attempt"""
        cfg = runner.load_config()
        if not cfg.attempt:
            print("No attempt in progress.")
            print("Use `tanco init` to start a new attempt.")
            return
        try:
            s = db.current_status(cfg.attempt)
            print("server:", s['server'])
            print("attempt:", cfg.attempt)
            print("challenge:", s['challenge'])
            print(f"state: {s['state']} {s['focus'] or ''}")
        except LookupError as e:
            print("attempt: ", cfg.attempt)
            print("attempt (or corresponding challenge) not found in database.")
            print("consider running `tanco recover`")

    def do_next(self, _arg):
        """Fetch the next test from the server."""
        # TODO:  double check that all tests pass and repo is clean

        cfg = runner.load_config()
        state = db.current_state(cfg.attempt)
        if state == 'done':
            print("You have already completed the challenge!")
            return
        elif state != 'start':
            # use `tanco show` to see if we already have the next test:
            self.do_show('-n')
            (attempt, known_tests) = self.result
            if known_tests:
                return

        # -- fetch the next test from the server
        tests = self.client.get_next(cfg.attempt)
        if not tests:
            print("You have completed the challenge!")
            db.set_attempt_state(cfg.uid, cfg.attempt, m.Transition.Done)
            # TODO: do something when you win
            return
        try:
            chid = db.challenge_from_attempt(cfg.attempt).id
            tx = db.begin()
            for t in tests:
                tx.execute("""
                    insert into tests (chid, name, head, body, grp, ord, ilines, olines)
                    values (?, ?, ?, ?, ?, ?, ?, ?)
                    """, [chid, t['name'], t['head'], t['body'], t['grp'], t['ord'],
                          t['ilines'], t['olines']])
            tx.commit()
            # have to do this second, or it'll transition to 'done'!!
            db.set_attempt_state(cfg.uid, cfg.attempt, m.Transition.Next)
        except sqlite3.IntegrityError as e:
            # this should not actually happen (because the 'show' call worked)
            # but just in case:
            print(e)

        self.do_show()

    @staticmethod
    def do_test(arg):
        """Run the tests."""
        runner.main(['tanco']+[x for x in arg.split(' ') if x != ''])

    @staticmethod
    def do_q(_arg):
        """Exit the shell."""
        return True

    @staticmethod
    def do_EOF(_arg):
        """Exit when ^D pressed or EOF reached"""
        return True


def show_help():
    print("tanco command line client")
    print()
    print("usage: tanco [command]")
    print()
    print("available commands:")
    print()
    data = [(meth[3:], getattr(TancoDriver, meth).__doc__)
            for meth in dir(TancoDriver)
            if meth.startswith('do_')]
    data.append(['shell', 'Run an interactive shell'])
    for cmd, doc in data:
        if cmd in ('EOF', 'help', 'q'): continue
        print("  %-16s:  %s" % (cmd, doc))


def main():
    db.ensure_sdb()
    d = TancoDriver()
    if "-q" in sys.argv:
        sys.argv.remove('-q')
        d.prompt = ""
    if len(sys.argv) > 1:
        cmd = sys.argv[1]
        if cmd == 'shell': d.cmdloop()
        else: d.onecmd(' '.join(sys.argv[1:]))
    else: show_help()
    sys.stderr.close()   # suppress warning on timeout when self-testing


if __name__ == '__main__':
    main()



"""
Implements a finite state machine to extract tanco
test descriptions from org files.

The org file format is described here:

  http://orgmode.org/worg/dev/org-syntax.html

However, there is no need to fully parse the outline
structure, since we are only interested in the actual
test cases, and these are clearly marked off by lines
in the following format:

   #+name: testname
   #+begin_src
    <test code winds up here>
   #+end_src

Test lines are simple sequences of lines in the
following format:

- Lines that begin with ">" are input lines.
  The test runner will send these to your program.

- Lines that begin with "'" are test descriptions,
  explaining the purpose of the test to the user.

- The "#" character indicates a comment, which is
  shown to the user but not part of the expected
  output.

- The "\" character can be used to escape any of
  these special characters (including itself).

"""
from collections import namedtuple
import os

from .model import TestDescription, Challenge

OldTestDescription = namedtuple("TestDescription", ['name', 'lines'])


class TestReaderStateMachine:
    """
    A simple finite state machine to extract tests from an outline.
    """
    def __init__(self):
        self.states = [ # NOTE: this is ordered and first field should == index
            (0, self.do_nothing),
            (1, self.on_test_code),
            (2, self.on_meta_line)]  # initial state for reading meta-data
        self.transitions = [
            (0, '#+name:', 0, self.on_test_name),
            (0, '#+begin_src', 1, self.on_begin_test),
            (1, '#+end_src',   0, self.on_end_test)]
        self.state = 2
        self.lineno = 0
        self.next_name = self.prev_name = ""
        self.test_names = []  # only for unique names
        self.tests = []       # collected (name, lines) descriptions
        self.challenge = Challenge()
        self.focus = None     # used to collect lines for current test

    # -- event handlers -----------------------------------------

    def on_line(self, line):
        self.lineno += 1
        match = [row[2:] for row in self.transitions 
                 if row[0] == self.state and line.startswith(row[1])]
        if match:  # match :: [( to_state, method )] of len 1
            self.state = match[0][0]
            match[0][1](line)
        else: self.states[self.state][1](line)

    def do_nothing(self, line):
        pass

    def on_meta_line(self, line):
        c = self.challenge
        if line.startswith('#+title:'):
            c.title = line.split(':')[1].strip()
        elif line.startswith('#+server:'):
            c.server = line.split(' ', 1)[1].strip()
        elif line.startswith('#+name:'):
            c.name = line.split(' ', 1)[1].strip()
        else:
            print(f'unexpected line {line!r} on line {self.lineno}')
            print('expect #+title:, #+server:, #+name: lines at start of file')
            exit()
        if c.title and c.server and c.name:
            self.state = 0

    def on_test_name(self, line):
        self.next_name = line.split(":")[1].strip()
        assert self.next_name not in self.test_names, (
            "duplicate name {0!r} on line {1}"
            .format(self.next_name, self.lineno))
        self.test_names.append(self.next_name)

    def on_begin_test(self, _line):
        assert self.next_name != self.prev_name, (
            "missing or duplicate name for test on line {0}"
            .format(self.lineno))
        self.tests.append(OldTestDescription(self.next_name, []))
        self.focus = self.tests[-1].lines
        self.prev_name = self.next_name

    def on_test_code(self, line):
        self.focus.append(line)

    def on_end_test(self, line):
        pass

    # -- main public interface ----------------------------------

    def parse(self, path):
        """
        Generates a sequence of TestDescription named tuples:
        Format is: (name:Str, lines:[Str])
        """
        for line in open(path): self.on_line(line)
        self.challenge.tests = [parse_test(x) for x in self.tests]
        return self.challenge


def parse_test(test):
    lines = test.lines
    while lines and lines[-1].strip() == "":
        lines.pop()
    opcodes = {
        'title': None,
        'doc': [],
        'in': [],
        'out': [],
    }
    for line in lines:
        if line.startswith('#'): continue
        if '#' in line:                # strip trailing comments
            line = line[:line.find('#')]
        sline = line.strip()
        if sline.startswith('='):      # test title
            opcodes['title'] = sline[2:]
        elif sline.startswith(':'):    # test description
            opcodes['doc'].append(sline)
        elif sline.startswith('>'):    # input to send
            opcodes['in'].append(sline[1:].lstrip())
        else:                          # expected output
            opcodes['out'].append(sline)
    step = TestDescription(
        name=test.name,
        head=opcodes['title'],
        body='\n'.join(opcodes['doc']),
        ilines=opcodes['in'],
        olines=opcodes['out'])
    return step


def read_challenge(path):
    return TestReaderStateMachine().parse(path)


def tests(path=None):
    """
    Convenience function to instantiate a TestReaderStateMachine
    and invoke .extract_tests on the given path.
    """
    if path is None:
        path = os.path.join(os.path.dirname(__file__), 'testplan.org')
    return read_challenge(path).tests


def main():
    if not os.path.exists('tests'): os.mkdir('tests')
    for i, test in enumerate(tests()):
        path = "tests/test{0:03}.txt".format(i)
        print("generating '{0}' in {1}"
              .format(test.name, path))
        io = open(path, 'w')
        for line in test.lines:
            io.write(line)
        io.close()


if __name__ == "__main__":
    main()



import requests
import os

from . import database as db
from . import model as m


class TancoClient:
    """
    Client for the Tanco API
    """

    def __init__(self, url=None):
        self.url = url or os.environ.get('TANCO_SERVER', 'https://tanco.tangentcode.com/')
        if not self.url.endswith('/'):
            self.url += '/'

    def post(self, url, data):
        url = url[1:] if url.startswith('/') else url
        res = requests.post(self.url + url, json=data)
        return res.json()

    def get_pre_token(self):
        return self.post('auth/pre', {})['token']

    def get_jwt(self, pre=None):
        return self.post('auth/jwt', {"pre": pre})['token']

    def list_challenges(self):
        res = requests.get(self.url + 'c.json')
        return res.json()

    def attempt(self, challenge_name):
        who = self.whoami()
        if not who:
            raise LookupError('You must be logged in to attempt a challenge.')
        res = requests.post(self.url + 'c/' + challenge_name + '/attempt',
                            json={'jwt': who['jwt']})
        return res.json()['aid']

    def whoami(self):
        res = db.query("""
            select u.id, u.username, t.jwt from tokens t, users u, servers s
            where t.uid=u.id and u.sid=s.id and s.url = ?
            """, [self.url])
        if len(res) > 1:
            raise LookupError('Multiple tokens found. This is a server database.')
        return res[0] if res else None

    def get_next(self, attempt):
        who = self.whoami()
        if not who:
            raise LookupError('You must be logged in to get next test.')
        return self.post('a/' + attempt + '/next', {'jwt': who['jwt']})

    def send_pass(self, attempt):
        who = self.whoami()
        if not who:
            raise LookupError('You must be logged in to send result.')
        db.set_attempt_state(who['id'], attempt, m.Transition.Pass)
        return self.post('a/' + attempt + '/pass', {
            'jwt': who['jwt']})

    def send_fail(self, attempt, test_name: str, result: m.TestResult):
        who = self.whoami()
        if not who:
            raise LookupError('You must be logged in to send result.')
        db.set_attempt_state(who['id'], attempt, m.Transition.Fail, test_name)
        return self.post('a/' + attempt + '/fail', {
            'jwt': who['jwt'],
            'test_name': test_name,
            'result': result.to_data()})

    def check_output(self, attempt: str, test_name: str, actual: [str]):
        who = self.whoami()
        if not who:
            raise LookupError('You must be logged in to check output.')
        res = self.post(f'a/{attempt}/check/{test_name}', {
            'jwt': who['jwt'],
            'actual': actual})
        return m.TestResult.from_data(res)
